;; Generated package description from slime.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "slime" "2.32snapshot0.20260505.41242" "Superior Lisp Interaction Mode for Emacs" '((emacs "24.3") (macrostep "0.9")) :commit "f866611800814f60f8b9845f9b4eddf617721383" :keywords '("languages" "lisp" "slime") :url "https://github.com/slime/slime")
